@@ -1,3 +1,4 @@
+# -*- coding: utf-8 -*-
 """
 merged implementation of the cache provider
 
@@ -60,10 +61,10 @@ class Cache(object):
 
     def warn(self, fmt, **args):
         from _pytest.warnings import _issue_warning_captured
-        from _pytest.warning_types import PytestWarning
+        from _pytest.warning_types import PytestCacheWarning
 
         _issue_warning_captured(
-            PytestWarning(fmt.format(**args) if args else fmt),
+            PytestCacheWarning(fmt.format(**args) if args else fmt),
             self._config.hook,
             stacklevel=3,
         )
@@ -157,18 +158,37 @@ class LFPlugin(object):
         self.active = any(config.getoption(key) for key in active_keys)
         self.lastfailed = config.cache.get("cache/lastfailed", {})
         self._previously_failed_count = None
-        self._no_failures_behavior = self.config.getoption("last_failed_no_failures")
+        self._report_status = None
+        self._skipped_files = 0  # count skipped files during collection due to --lf
+
+    def last_failed_paths(self):
+        """Returns a set with all Paths()s of the previously failed nodeids (cached).
+        """
+        try:
+            return self._last_failed_paths
+        except AttributeError:
+            rootpath = Path(self.config.rootdir)
+            result = {rootpath / nodeid.split("::")[0] for nodeid in self.lastfailed}
+            result = {x for x in result if x.exists()}
+            self._last_failed_paths = result
+            return result
+
+    def pytest_ignore_collect(self, path):
+        """
+        Ignore this file path if we are in --lf mode and it is not in the list of
+        previously failed files.
+        """
+        if self.active and self.config.getoption("lf") and path.isfile():
+            last_failed_paths = self.last_failed_paths()
+            if last_failed_paths:
+                skip_it = Path(path) not in self.last_failed_paths()
+                if skip_it:
+                    self._skipped_files += 1
+                return skip_it
 
     def pytest_report_collectionfinish(self):
         if self.active and self.config.getoption("verbose") >= 0:
-            if not self._previously_failed_count:
-                return None
-            noun = "failure" if self._previously_failed_count == 1 else "failures"
-            suffix = " first" if self.config.getoption("failedfirst") else ""
-            mode = "rerun previous {count} {noun}{suffix}".format(
-                count=self._previously_failed_count, suffix=suffix, noun=noun
-            )
-            return "run-last-failure: %s" % mode
+            return "run-last-failure: %s" % self._report_status
 
     def pytest_runtest_logreport(self, report):
         if (report.when == "call" and report.passed) or report.skipped:
@@ -186,28 +206,51 @@ class LFPlugin(object):
             self.lastfailed[report.nodeid] = True
 
     def pytest_collection_modifyitems(self, session, config, items):
-        if self.active:
-            if self.lastfailed:
-                previously_failed = []
-                previously_passed = []
-                for item in items:
-                    if item.nodeid in self.lastfailed:
-                        previously_failed.append(item)
-                    else:
-                        previously_passed.append(item)
-                self._previously_failed_count = len(previously_failed)
-                if not previously_failed:
-                    # running a subset of all tests with recorded failures outside
-                    # of the set of tests currently executing
-                    return
+        if not self.active:
+            return
+
+        if self.lastfailed:
+            previously_failed = []
+            previously_passed = []
+            for item in items:
+                if item.nodeid in self.lastfailed:
+                    previously_failed.append(item)
+                else:
+                    previously_passed.append(item)
+            self._previously_failed_count = len(previously_failed)
+
+            if not previously_failed:
+                # Running a subset of all tests with recorded failures
+                # only outside of it.
+                self._report_status = "%d known failures not in selected tests" % (
+                    len(self.lastfailed),
+                )
+            else:
                 if self.config.getoption("lf"):
                     items[:] = previously_failed
                     config.hook.pytest_deselected(items=previously_passed)
-                else:
+                else:  # --failedfirst
                     items[:] = previously_failed + previously_passed
-            elif self._no_failures_behavior == "none":
+
+                noun = "failure" if self._previously_failed_count == 1 else "failures"
+                suffix = " first" if self.config.getoption("failedfirst") else ""
+                self._report_status = "rerun previous {count} {noun}{suffix}".format(
+                    count=self._previously_failed_count, suffix=suffix, noun=noun
+                )
+
+            if self._skipped_files > 0:
+                files_noun = "file" if self._skipped_files == 1 else "files"
+                self._report_status += " (skipped {files} {files_noun})".format(
+                    files=self._skipped_files, files_noun=files_noun
+                )
+        else:
+            self._report_status = "no previously failed tests, "
+            if self.config.getoption("last_failed_no_failures") == "none":
+                self._report_status += "deselecting all items."
                 config.hook.pytest_deselected(items=items)
                 items[:] = []
+            else:
+                self._report_status += "not deselecting items."
 
     def pytest_sessionfinish(self, session):
         config = self.config
@@ -282,9 +325,13 @@ def pytest_addoption(parser):
     )
     group.addoption(
         "--cache-show",
-        action="store_true",
+        action="append",
+        nargs="?",
         dest="cacheshow",
-        help="show cache contents, don't perform collection or tests",
+        help=(
+            "show cache contents, don't perform collection or tests. "
+            "Optional argument: glob (default: '*')."
+        ),
     )
     group.addoption(
         "--cache-clear",
@@ -303,8 +350,7 @@ def pytest_addoption(parser):
         dest="last_failed_no_failures",
         choices=("all", "none"),
         default="all",
-        help="change the behavior when no test failed in the last run or no "
-        "information about the last failures was found in the cache",
+        help="which tests to run with no previously (known) failures.",
     )
 
 
@@ -360,11 +406,16 @@ def cacheshow(config, session):
     if not config.cache._cachedir.is_dir():
         tw.line("cache is empty")
         return 0
+
+    glob = config.option.cacheshow[0]
+    if glob is None:
+        glob = "*"
+
     dummy = object()
     basedir = config.cache._cachedir
     vdir = basedir / "v"
-    tw.sep("-", "cache values")
-    for valpath in sorted(x for x in vdir.rglob("*") if x.is_file()):
+    tw.sep("-", "cache values for %r" % glob)
+    for valpath in sorted(x for x in vdir.rglob(glob) if x.is_file()):
         key = valpath.relative_to(vdir)
         val = config.cache.get(key, dummy)
         if val is dummy:
@@ -376,8 +427,8 @@ def cacheshow(config, session):
 
     ddir = basedir / "d"
     if ddir.is_dir():
-        contents = sorted(ddir.rglob("*"))
-        tw.sep("-", "cache directories")
+        contents = sorted(ddir.rglob(glob))
+        tw.sep("-", "cache directories for %r" % glob)
         for p in contents:
             # if p.check(dir=1):
             #    print("%s/" % p.relto(basedir))
